# -*- coding: utf-8 -*-
"""
    wakatime.projects.git
    ~~~~~~~~~~~~~~~~~~~~~

    Information about the git project for a given file.

    :copyright: (c) 2013 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

import logging
import os
import re
import sys

from .base import BaseProject
from ..compat import u, open


log = logging.getLogger('WakaTime')


class Git(BaseProject):
    _submodule = None
    _project_name = None
    _head_file = None
    _project_folder = None

    def process(self):
        return self._find_git_config_file(self.path)

    def name(self):
        return u(self._project_name) if self._project_name else None

    def branch(self):
        head = self._head_file
        if head:
            line = self._first_line_of_file(head)
            if line is not None:
                return self._get_branch_from_head_file(line)
        return u('master')

    def folder(self):
        return self._project_folder

    def _find_git_config_file(self, path):
        path = os.path.realpath(path)
        if os.path.isfile(path):
            path = os.path.split(path)[0]
        if os.path.isfile(os.path.join(path, '.git', 'config')):
            self._project_name = os.path.basename(path)
            self._head_file = os.path.join(path, '.git', 'HEAD')
            self._project_folder = path
            return True

        link_path = self._path_from_gitdir_link_file(path)
        if link_path:

            # first check if this is a worktree
            if self._is_worktree(link_path):
                self._project_name = self._project_from_worktree(link_path)
                self._head_file = os.path.join(link_path, 'HEAD')
                self._project_folder = path
                return True

            # next check if this is a submodule
            if self._submodules_supported_for_path(path):
                self._project_name = os.path.basename(path)
                self._head_file = os.path.join(link_path, 'HEAD')
                self._project_folder = path
                return True

        split_path = os.path.split(path)
        if split_path[1] == '':
            return False
        return self._find_git_config_file(split_path[0])

    def _get_branch_from_head_file(self, line):
        if u(line.strip()).startswith('ref: '):
            return u(line.strip().rsplit('/', 1)[-1])
        return None

    def _submodules_supported_for_path(self, path):
        if not self._configs:
            return True

        disabled = self._configs.get('submodules_disabled')

        if not disabled or disabled.strip().lower() == 'false':
            return True
        if disabled.strip().lower() == 'true':
            return False

        for pattern in disabled.split("\n"):
            if pattern.strip():
                try:
                    compiled = re.compile(pattern, re.IGNORECASE)
                    if compiled.search(path):
                        return False
                except re.error as ex:
                    log.warning(u('Regex error ({msg}) for disable git submodules pattern: {pattern}').format(
                        msg=u(ex),
                        pattern=u(pattern),
                    ))

        return True

    def _is_worktree(self, link_path):
        return os.path.basename(os.path.dirname(link_path)) == 'worktrees'

    def _path_from_gitdir_link_file(self, path):
        link = os.path.join(path, '.git')
        if not os.path.isfile(link):
            return None

        line = self._first_line_of_file(link)
        if line is not None:
            return self._path_from_gitdir_string(path, line)

        return None

    def _path_from_gitdir_string(self, path, line):
        if line.startswith('gitdir: '):
            subpath = line[len('gitdir: '):].strip()
            if os.path.isfile(os.path.join(path, subpath, 'HEAD')):
                return os.path.realpath(os.path.join(path, subpath))

        return None

    def _project_from_worktree(self, link_path):
        commondir = os.path.join(link_path, 'commondir')
        if os.path.isfile(commondir):
            line = self._first_line_of_file(commondir)
            if line:
                gitdir = os.path.abspath(os.path.join(link_path, line))
                if os.path.basename(gitdir) == '.git':
                    return os.path.basename(os.path.dirname(gitdir))

        return None

    def _first_line_of_file(self, filepath):
        try:
            with open(filepath, 'r', encoding='utf-8') as fh:
                return fh.readline().strip()
        except UnicodeDecodeError:
            pass
        except IOError:
            pass
        try:
            with open(filepath, 'r', encoding=sys.getfilesystemencoding()) as fh:
                return fh.readline().strip()
        except:
            log.traceback(logging.WARNING)

        return None



# -*- coding: utf-8 -*-
"""
    wakatime.dependencies.objective
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Parse dependencies from Objective-C and Swift code.

    :copyright: (c) 2018 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

import re

from . import TokenParser


class SwiftParser(TokenParser):
    state = None
    exclude = [
        r'^foundation$',
    ]

    def parse(self):
        for index, token, content in self.tokens:
            self._process_token(token, content)
        return self.dependencies

    def _process_token(self, token, content):
        if self.partial(token) == 'Declaration':
            self._process_declaration(token, content)
        elif self.partial(token) == 'Class':
            self._process_class(token, content)
        else:
            self._process_other(token, content)

    def _process_declaration(self, token, content):
        if self.state is None:
            self.state = content

    def _process_class(self, token, content):
        if self.state == 'import':
            self.append(content)
        self.state = None

    def _process_other(self, token, content):
        pass


class ObjectiveCParser(TokenParser):
    state = None
    extension = re.compile(r'\.[mh]$')

    def parse(self):
        for index, token, content in self.tokens:
            self._process_token(token, content)
        return self.dependencies

    def _process_token(self, token, content):
        if self.partial(token) == 'Preproc':
            self._process_preproc(token, content)
        else:
            self._process_other(token, content)

    def _process_preproc(self, token, content):
        if self.state:
            self._process_import(token, content)

        self.state = content

    def _process_import(self, token, content):
        if self.state == '#' and content.startswith('import '):
            self.append(self._format(content))
        self.state = None

    def _process_other(self, token, content):
        pass

    def _format(self, content):
        content = content.strip().lstrip('import ').strip()
        content = content.strip('"').strip("'").strip()
        content = content.strip('<').strip('>').strip()
        content = content.split('/')[0]
        content = self.extension.sub('', content, count=1)
        return content



# -*- coding: utf-8 -*-
"""
    wakatime.project
    ~~~~~~~~~~~~~~~~

    Returns a project for the given file.

    :copyright: (c) 2013 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

import os
import logging
import random

from .compat import open
from .projects.git import Git
from .projects.mercurial import Mercurial
from .projects.projectfile import ProjectFile
from .projects.projectmap import ProjectMap
from .projects.subversion import Subversion


log = logging.getLogger('WakaTime')


# List of plugin classes to find a project for the current file path.
CONFIG_PLUGINS = [
    ProjectFile,
    ProjectMap,
]
REV_CONTROL_PLUGINS = [
    Git,
    Mercurial,
    Subversion,
]


def get_project_info(configs, heartbeat, data):
    """Find the current project and branch.

    First looks for a .wakatime-project file. Second, uses the --project arg.
    Third, uses the folder name from a revision control repository. Last, uses
    the --alternate-project arg.

    Returns a project, branch tuple.
    """

    project_name, branch_name = heartbeat.project, heartbeat.branch

    if heartbeat.type != 'file':
        project_name = project_name or heartbeat.args.project or heartbeat.args.alternate_project
        return project_name, branch_name

    if project_name is None or branch_name is None:

        for plugin_cls in CONFIG_PLUGINS:

            plugin_name = plugin_cls.__name__.lower()
            plugin_configs = get_configs_for_plugin(plugin_name, configs)

            project = plugin_cls(heartbeat.entity, configs=plugin_configs)
            if project.process():
                project_name = project_name or project.name()
                branch_name = project.branch()
                break

    if project_name is None:
        project_name = data.get('project') or heartbeat.args.project

    hide_project = heartbeat.should_obfuscate_project()
    if hide_project and project_name is not None:
        return project_name, None

    if project_name is None or branch_name is None:

        for plugin_cls in REV_CONTROL_PLUGINS:

            plugin_name = plugin_cls.__name__.lower()
            plugin_configs = get_configs_for_plugin(plugin_name, configs)

            project = plugin_cls(heartbeat.entity, configs=plugin_configs)
            if project.process():
                project_name = project_name or project.name()
                branch_name = branch_name or project.branch()
                if hide_project:
                    branch_name = None
                    project_name = generate_project_name()
                    project_file = os.path.join(project.folder(), '.wakatime-project')
                    try:
                        with open(project_file, 'w') as fh:
                            fh.write(project_name)
                    except IOError:
                        project_name = None
                break

    if project_name is None and not hide_project:
        project_name = data.get('alternate_project') or heartbeat.args.alternate_project

    return project_name, branch_name


def get_configs_for_plugin(plugin_name, configs):
    if configs and configs.has_section(plugin_name):
        return dict(configs.items(plugin_name))
    return None


def generate_project_name():
    """Generates a random project name."""

    adjectives = [
        'aged', 'ancient', 'autumn', 'billowing', 'bitter', 'black', 'blue', 'bold',
        'broad', 'broken', 'calm', 'cold', 'cool', 'crimson', 'curly', 'damp',
        'dark', 'dawn', 'delicate', 'divine', 'dry', 'empty', 'falling', 'fancy',
        'flat', 'floral', 'fragrant', 'frosty', 'gentle', 'green', 'hidden', 'holy',
        'icy', 'jolly', 'late', 'lingering', 'little', 'lively', 'long', 'lucky',
        'misty', 'morning', 'muddy', 'mute', 'nameless', 'noisy', 'odd', 'old',
        'orange', 'patient', 'plain', 'polished', 'proud', 'purple', 'quiet', 'rapid',
        'raspy', 'red', 'restless', 'rough', 'round', 'royal', 'shiny', 'shrill',
        'shy', 'silent', 'small', 'snowy', 'soft', 'solitary', 'sparkling', 'spring',
        'square', 'steep', 'still', 'summer', 'super', 'sweet', 'throbbing', 'tight',
        'tiny', 'twilight', 'wandering', 'weathered', 'white', 'wild', 'winter', 'wispy',
        'withered', 'yellow', 'young'
    ]
    nouns = [
        'art', 'band', 'bar', 'base', 'bird', 'block', 'boat', 'bonus',
        'bread', 'breeze', 'brook', 'bush', 'butterfly', 'cake', 'cell', 'cherry',
        'cloud', 'credit', 'darkness', 'dawn', 'dew', 'disk', 'dream', 'dust',
        'feather', 'field', 'fire', 'firefly', 'flower', 'fog', 'forest', 'frog',
        'frost', 'glade', 'glitter', 'grass', 'hall', 'hat', 'haze', 'heart',
        'hill', 'king', 'lab', 'lake', 'leaf', 'limit', 'math', 'meadow',
        'mode', 'moon', 'morning', 'mountain', 'mouse', 'mud', 'night', 'paper',
        'pine', 'poetry', 'pond', 'queen', 'rain', 'recipe', 'resonance', 'rice',
        'river', 'salad', 'scene', 'sea', 'shadow', 'shape', 'silence', 'sky',
        'smoke', 'snow', 'snowflake', 'sound', 'star', 'sun', 'sun', 'sunset',
        'surf', 'term', 'thunder', 'tooth', 'tree', 'truth', 'union', 'unit',
        'violet', 'voice', 'water', 'waterfall', 'wave', 'wildflower', 'wind', 'wood'
    ]
    numbers = [str(x) for x in range(10)]
    return ' '.join([
        random.choice(adjectives).capitalize(),
        random.choice(nouns).capitalize(),
        random.choice(numbers) + random.choice(numbers),
    ])



# -*- coding: utf-8 -*-
"""
    wakatime.dependencies.haskell
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Parse dependencies from Haskell code.

    :copyright: (c) 2018 Alan Hamlett.
    :license: BSD, see LICENSE for more details.
"""

from . import TokenParser


class HaskellParser(TokenParser):
    state = None

    def parse(self):
        for index, token, content in self.tokens:
            self._process_token(token, content)
        return self.dependencies

    def _process_token(self, token, content):
        if self.partial(token) == 'Reserved':
            self._process_reserved(token, content)
        elif self.partial(token) == 'Namespace':
            self._process_namespace(token, content)
        elif self.partial(token) == 'Keyword':
            self._process_keyword(token, content)
        elif self.partial(token) == 'Text':
            self._process_text(token, content)
        else:
            self._process_other(token, content)

    def _process_reserved(self, token, content):
        self.state = content.strip()

    def _process_namespace(self, token, content):
        if self.state == 'import':
            self.append(self._format(content))

    def _process_keyword(self, token, content):
        if self.state != 'import' or content.strip() != 'qualified':
            self.state = None

    def _process_text(self, token, content):
        pass

    def _process_other(self, token, content):
        self.state = None

    def _format(self, content):
        return content.strip().split('.')[0].strip()
